import ast
import sys
import os
import re
import codecs
from py_compile import PyCompileError
import fnmatch
import signal
from nltk import word_tokenize

# Class that visits all Class nodes in given tree and stores them into array
class ClassNodeVisitor(ast.NodeVisitor):

    def __init__(self):
        self.class_nodes = []

    def visit_ClassDef(self, node):
        self.class_nodes.append(node)
        self.generic_visit(node)

# Class that visits all FunctionDef nodes in given tree and stores them into array
class FunctionNodeVisitor(ast.NodeVisitor):

    def __init__(self):
        self.function_nodes = []

    # don't visit any function nodes within classes
    def visit_ClassDef(self, node):
        pass

    def visit_FunctionDef(self, node):
        self.function_nodes.append(node)
        self.generic_visit(node)

# Class that visits all method call (Call) nodes in the given tree,
# stores them and their names into array
class CallVisitor(ast.NodeVisitor):

    def __init__(self):
        self.call_nodes = []

    def visit_Call(self, node):
        callee_name = None
        caller_name = None
        for (x, y) in ast.iter_fields(node):
            if caller_name is not None and callee_name is not None:
                break
            if isinstance(y, ast.Attribute):
                for (field_x, field_y) in ast.iter_fields(y):
                    if field_x == "attr":
                        callee_name = field_y
                    elif field_x == "value":
                        for (child_field_x, child_field_y) in ast.iter_fields(field_y):
                            if child_field_x == "attr" or child_field_x == "id":
                                caller_name = child_field_y
                                break
        if caller_name is not None and caller_name is not "self" and callee_name is not None:
            self.call_nodes.append((caller_name, callee_name, node))
        self.generic_visit(node)

def flatten(list):
    new_list = []
    for item in list:
        if type(item) == type([]):
            new_list.extend(flatten(item))
        else:
            new_list.append(item)
    return new_list

def tokenize(sentence):
    try:
        tokens = word_tokenize(sentence)
        for token in tokens:
            if "." in token:
                subtokens = token.split(".")
                subtokens = list(filter(None, subtokens))
                index = tokens.index(token)
                tokens.remove(token)
                tokens.insert(index, subtokens)
        tokens = flatten(tokens)
        return " ".join(tokens)
    except (TypeError, UnicodeDecodeError):
        return ""

#clears comments from code line
def strip_comments(line):
    if "'''" in line:
        beginCommentIndex = line.index("'''")
        endCommentIndex = line.rindex("'''")
        if(beginCommentIndex != endCommentIndex and endCommentIndex > beginCommentIndex):
            return line[0:beginCommentIndex] + line[endCommentIndex+3:len(line)]
    elif "#" in line:
        beginCommentIndex = line.index("#")
        return line[0:beginCommentIndex]
    else:
        return line

def collect_calls(ast_node, file_lines):
    input_lines = []
    eval_lines = []
    node_def_line_num = ast_node.lineno
    call_visitor = CallVisitor()
    try:
        call_visitor.visit(ast_node)
    except RuntimeError:
        return [], []
    call_nodes = call_visitor.call_nodes
    for (caller_name, callee_name, node) in call_nodes:
        input_line = ""
        # node.lineno is 1-indexed
        call_line_num = node.lineno - 1

        # relevant scope for every variable are lines from enclosing function definition to actual line where variable calls a method
        relevant_lines = file_lines[node_def_line_num:call_line_num]
        for code_line in relevant_lines:
            # match exact variable name
            if re.search(r'\b' + caller_name + r'\b', code_line):
                input_line += tokenize(strip_comments(code_line)) + " "
        code_line = file_lines[call_line_num]

        # get part of the code line before method call i.e. result = node.get(...) returns result = node
        prefix_code_line = code_line[0:(code_line.index(caller_name+".") + len(caller_name))]

        input_line += tokenize(prefix_code_line)
        input_line += "\n"
        #filter out all self to don't confuse the model
        input_line = input_line.replace("self ", "")
        input_lines.append(input_line)
        eval_lines.append(callee_name + "\n")
    return input_lines, eval_lines

def process_file(input_file):
    try:
        with codecs.open(input_file, 'r', encoding='utf-8', errors='ignore') as f:
            input_lines = []
            eval_lines = []

            file = f.read()
            lines = file.splitlines()
            try:
                tree = ast.parse(file)

                # visit every class node and extract all calls from it
                class_visitor = ClassNodeVisitor()
                try:
                    class_visitor.visit(tree)
                    class_nodes = class_visitor.class_nodes
                    for class_node in class_nodes:
                        class_input_lines, class_output_lines = collect_calls(class_node, lines)
                        input_lines.extend(class_input_lines)
                        eval_lines.extend(class_output_lines)
                except RuntimeError as runtime_error:
                    print("Error processing file: ", input_file + ": ", runtime_error)
                    return [], []

                # visit every function node and extract all calls from it
                function_visitor = FunctionNodeVisitor()
                try:
                    function_visitor.visit(tree)
                    function_nodes = function_visitor.function_nodes

                    for function_node in function_nodes:
                        func_def_input_lines, func_def_output_lines = collect_calls(function_node, lines)
                        input_lines.extend(func_def_input_lines)
                        eval_lines.extend(func_def_output_lines)
                except RuntimeError as runtime_error:
                    print("Error processing file: ", input_file + ": ", runtime_error)
                    return [], []

            except (PyCompileError, SyntaxError, ValueError, TypeError):
                print("Failed parsing file:", input_file)
                pass

            return input_lines, eval_lines
    except IOError as ioError:
        print("I/O error processing file", input_file, ioError.strerror)
        return [], []

#recursively searches and returns filenames in given path with given extension
def retrieve_files(path, extension):
	matches = []
	for root, dirnames, filenames in os.walk(path):
		for filename in fnmatch.filter(filenames, '*' + extension):
			matches.append(os.path.join(root, filename))
	return matches

def main(dir):
    src_file = open("tokenized_input.txt", "w+")
    eval_file = open("eval_input.txt", "w+")
    code_files = retrieve_files(dir, ".py")
    print("Total " + str(len(code_files)) + " files to process")
    counter = 0
    for code_file in code_files:
        print("Processing " + str(counter) + " file out of " + str(len(code_files)))
        counter += 1
        signal.alarm(30)
        try:
            input_lines, eval_lines = process_file(code_file)
            if (len(input_lines) == len(eval_lines) and len(input_lines) > 0):
                for i in range(0, len(input_lines)):
                    src_file.write(input_lines[i])
                    eval_file.write(eval_lines[i])
            else:
                if (len(input_lines) > 0):
                    print("Number of input file lines doesn't match number of evaluation lines!")
        except TimeoutError:
            print("Processing file " + code_file + " timed out!")
        signal.alarm(0)
    src_file.close()
    eval_file.close()

#########################################
#          TIMEOUT ROUTINE
#########################################
class TimeoutError(Exception):
    pass

def timeout_handler(num, stack):
    raise TimeoutError()

if __name__ == "__main__":
    signal.signal(signal.SIGALRM, timeout_handler)
    main(sys.argv[1])