import sys
import os
import ast
import codecs
import fnmatch
from py_compile import PyCompileError
from nltk import word_tokenize

# Class that visits all method call (Call) nodes in the given tree,
# stores them and their names into array
class CallVisitor(ast.NodeVisitor):

    def __init__(self):
        self.call_nodes = []

    def visit_Call(self, node):
        callee_name = None
        caller_name = None
        for (x, y) in ast.iter_fields(node):
            if caller_name is not None and callee_name is not None:
                break
            if isinstance(y, ast.Attribute):
                for (field_x, field_y) in ast.iter_fields(y):
                    if field_x == "attr":
                        callee_name = field_y
                    elif field_x == "value":
                        for (child_field_x, child_field_y) in ast.iter_fields(field_y):
                            if child_field_x == "attr" or child_field_x == "id":
                                caller_name = child_field_y
                                break
        if caller_name is not None and caller_name is not "self" and callee_name is not None:
            self.call_nodes.append((caller_name, callee_name, node))
        self.generic_visit(node)

def flatten(list):
    new_list = []
    for item in list:
        if type(item) == type([]):
            new_list.extend(flatten(item))
        else:
            new_list.append(item)
    return new_list

def tokenize(sentence):
    tokens = word_tokenize(sentence)
    for token in tokens:
        if "." in token:
            subtokens = token.split(".")
            subtokens = list(filter(None, subtokens))
            index = tokens.index(token)
            tokens.remove(token)
            tokens.insert(index, subtokens)
    tokens = flatten(tokens)
    return " ".join(tokens)

def process_file(input_file):
    try:
        with codecs.open(input_file, 'r', encoding='utf-8', errors='ignore') as f:
            input_lines = []
            eval_lines = []

            file = f.read()
            lines = file.splitlines()
            try:
                tree = ast.parse(file)

                # visit all call nodes in the class
                call_visitor = CallVisitor()
                try:
                    call_visitor.visit(tree)
                    call_nodes = call_visitor.call_nodes
                    for (caller_name, callee_name, node) in call_nodes:
                        # node.lineno is 1-indexed
                        call_line_num = node.lineno - 1
                        code_line = lines[call_line_num]
                        # get part of the code line before method call i.e. result = node.get(...) returns result = node
                        prefix_code_line = code_line[0:(code_line.index(caller_name + ".") + len(caller_name))]
                        input_line = tokenize(prefix_code_line)
                        # filter out all self to don't confuse the model
                        input_line = input_line.replace("self ", "")
                        input_line += "\n"
                        input_lines.append(input_line)
                        eval_lines.append(callee_name + "\n")
                except RuntimeError as runtime_error:
                    print("Error processing file: ", input_file + ": ", runtime_error)
                    return [], []

            except (PyCompileError, SyntaxError, ValueError, TypeError):
                print("Failed parsing file:", input_file)
                pass

            return input_lines, eval_lines
    except IOError as ioError:
        print("I/O error processing file", input_file, ioError.strerror)
        return [], []

#recursively searches and returns filenames in given path with given extension
def retrieve_files(path, extension):
	matches = []
	for root, dirnames, filenames in os.walk(path):
		for filename in fnmatch.filter(filenames, '*' + extension):
			matches.append(os.path.join(root, filename))
	return matches

def main(dir):
    src_file = open("tokenized_input.txt", "w+")
    eval_file = open("eval_input.txt", "w+")
    code_files = retrieve_files(dir, ".py")
    print("Total " + str(len(code_files)) + " files to process")
    counter = 0
    for code_file in code_files:
        print("Processing " + str(counter) + " file out of " + str(len(code_files)))
        src_lines, eval_lines  = process_file(code_file)
        counter += 1
        if len(src_lines) == len(eval_lines) and len(src_lines) > 0:
            for i in range(0, len(src_lines)):
                src_file.write(src_lines[i])
                eval_file.write(eval_lines[i])
        else:
            if (len(src_lines) > 0):
                print("Number of input file lines doesn't match number of evaluation lines!")
    src_file.close()
    eval_file.close()

if __name__ == "__main__":
    main(sys.argv[1])