import ast
import os
import codecs
from py_compile import PyCompileError
import sys
import fnmatch

# how many characters before called method name to take (if file contains less than 1000, takes whatever is available)
MAX_CHAR_COUNT = 1000

#replacement for newline character
NEWLINE_CHAR = "\u00ff"
#replacement for tab character
TAB_CHAR = "\uf8ff"
#replacement for space character
SPACE_CHAR = "\uf8fe"

# Class that visits all method call (Call) nodes in the given tree,
# stores them and their names into array
class CallVisitor(ast.NodeVisitor):

    def __init__(self):
        self.call_nodes = []

    def visit_Call(self, node):
        callee_name = None
        for (x, y) in ast.iter_fields(node):
            if callee_name is not None:
                break
            if isinstance(y, ast.Attribute):
                for (field_x, field_y) in ast.iter_fields(y):
                    if field_x == "attr":
                        callee_name = field_y
                        self.call_nodes.append((callee_name, node))
                        break
        self.generic_visit(node)

def replace_characters(line):
    line = line.replace("\n", NEWLINE_CHAR)
    line = line.replace(" ", SPACE_CHAR)
    line = line.replace("\t", TAB_CHAR)
    return " ".join(line)

def process_file(input_file):
    try:
        with codecs.open(input_file, 'r', encoding='utf-8', errors='ignore') as f:
            input_lines = []
            eval_lines = []

            file = f.read()
            lines = file.splitlines()
            try:
                tree = ast.parse(file)

                call_visitor = CallVisitor()
                call_visitor.visit(tree)

                call_nodes = call_visitor.call_nodes
                for (callee_name, node) in call_nodes:
                    # node.lineno is 1-indexed
                    call_line_num = node.lineno - 1
                    code_line = lines[call_line_num]

                    char_count = 0
                    result = []
                    # get part of the code line before method call i.e. result = node.get(...) returns result = node
                    prefix_code_line = code_line[:code_line.index(callee_name)]
                    char_count += len(prefix_code_line)
                    if char_count >= MAX_CHAR_COUNT:
                        # ensure that entire words are taken and length of char sequence isn't above CHAR_COUNT
                        num_chars_to_cutoff = char_count - MAX_CHAR_COUNT
                        string_part = prefix_code_line
                        if num_chars_to_cutoff > 0:
                            string_part = prefix_code_line[num_chars_to_cutoff:]
                            try:
                                first_space_idx = string_part.index(" ")
                                string_part = string_part[(first_space_idx + 1):]
                            except ValueError:
                                pass
                        if len(string_part) < MAX_CHAR_COUNT:
                            result_line = replace_characters(string_part)
                            input_lines.append(result_line + "\n")
                            eval_lines.append(callee_name + "\n")
                    else:
                        result.append(prefix_code_line)
                        line_index = call_line_num - 1
                        while line_index >= 0:
                            prev_code_line = lines[line_index]
                            prev_code_line += "\n"
                            char_count += len(prev_code_line)
                            if char_count > MAX_CHAR_COUNT:
                                # ensure that entire words are taken and length of char sequence isn't above CHAR_COUNT
                                num_chars_to_cutoff = char_count - MAX_CHAR_COUNT
                                string_part = prev_code_line[num_chars_to_cutoff:]
                                try:
                                    first_space_idx = string_part.index(" ")
                                    string_part = string_part[(first_space_idx+1):]
                                except ValueError:
                                    pass
                                result.append(string_part)
                                break
                            else:
                                result.append(prev_code_line)
                            line_index -= 1

                        result.reverse()
                        single_line = "".join(result)
                        input_line = replace_characters(single_line)
                        input_lines.append(input_line + "\n")
                        eval_lines.append(callee_name + "\n")
            except (PyCompileError, SyntaxError, ValueError, TypeError, RuntimeError):
                print("Failed parsing file:", input_file)
                pass

            return input_lines, eval_lines
    except IOError as ioError:
        print("I/O error processing file", input_file, ioError.strerror)
        return [], []

#recursively searches and returns filenames in given path with given extension
def retrieve_files(path, extension):
	matches = []
	for root, dirnames, filenames in os.walk(path):
		for filename in fnmatch.filter(filenames, '*' + extension):
			matches.append(os.path.join(root, filename))
	return matches

def main(dir):
    src_file = open("tokenized_input.txt", "w+")
    eval_file = open("eval_input.txt", "w+")
    code_files = retrieve_files(dir, ".py")
    print("Total " + str(len(code_files)) + " files to process")
    counter = 0
    for code_file in code_files:
        print("Processed " + str(counter) + " files out of " + str(len(code_files)))
        counter += 1
        input_lines, eval_lines = process_file(code_file)
        if(len(input_lines) == len(eval_lines) and len(input_lines) > 0):
            for i in range(0, len(input_lines)):
                src_file.write(input_lines[i])
                eval_file.write(eval_lines[i])
        else:
            if(len(input_lines) > 0):
                print("Number of input file lines doesn't match number of evaluation lines!")
    src_file.close()
    eval_file.close()

if __name__ == "__main__":
    main(sys.argv[1])