#!/bin/bash

if [[ "$1" = "--help" ]]; then
  echo 'usage: ./train.bash [--test] EPOCHS PARTS GPUS DATA_PATH'
  echo 'example: ./train.bash 7 8 "1,2,3" ./test-data'
  exit 0
fi

# When --test is passed, th commands are only echoed and dummy model files are created
[[ "$1" = "--test" ]] && TESTING="true" && shift

REQ_ARGS=4
[[ "$#" -ne "$REQ_ARGS" ]] && echo "Requires exactly "$REQ_ARGS" arguments (see --help)" && exit 1

NUM_EPOCHS="$1"
NUM_PARTS="$2"
GPU_ID="$3"
DATA_PATH="$4"
LOG_PATH="$DATA_PATH/th-train.log"

[[ ! -d "$DATA_PATH" ]] && echo "$DATA_PATH is not a directory" && exit 2

echo "Epochs: $NUM_EPOCHS | parts: $NUM_PARTS | GPUs: $GPU_ID | data: $DATA_PATH"

for epoch in $(seq 1 $NUM_EPOCHS); do
  for part in $(seq 1 $NUM_PARTS); do

    echo "epoch $epoch, part $part - running train.lua"

    # skip to next epoch/part if current model already exists
    current_model=$(find "$DATA_PATH" -name "*model_part${part}_epoch${epoch}_*.t7")
    if [[ -n "$current_model" ]]; then
      echo "current model file $current_model already exists, skipping"
      continue
    fi

    data_file="$DATA_PATH/data-$part-train.t7"
    [[ ! "$TESTING" = true ]] && [[ ! -f "$data_file" ]] && echo "$data_file is not a file" && exit 3

    if [[ "$epoch"  -eq 1 ]] && [[ "$part" -eq 1 ]]; then

      if [[ "$TESTING" = true ]]; then
        echo th train.lua -data "$data_file" -save_model "$DATA_PATH/model_part${part}" -start_epoch "$epoch" -end_epoch "$epoch" -gpuid "$GPU_ID"
        touch "$DATA_PATH/model_part${part}_epoch${epoch}_1.23.t7"
      else
        th train.lua -data "$DATA_PATH/data-$part-train.t7" -save_model "$DATA_PATH/model_part${part}" -start_epoch "$epoch" -end_epoch "$epoch" -gpuid "$GPU_ID" >> "$LOG_PATH" 2>&1
      fi

    else

      # previous model part: last if current part is 1, else part-1
      prev_part="$((part-1))"
      if [[ "$part" -eq 1 ]]; then prev_part="$NUM_PARTS"; fi
      # previous model epoch: previous if current part is 1, else same epoch
      prev_epoch="$epoch"
      if [[ "$part" -eq 1 ]]; then prev_epoch="$((epoch-1))"; fi
      prev_model=$(find "$DATA_PATH" -name "*model_part${prev_part}_epoch${prev_epoch}_*.t7")

      if [[ "$TESTING" = true ]]; then
        echo th train.lua -data "$data_file" -train_from "$prev_model" -save_model $DATA_PATH/model_part${part} -start_epoch "$epoch" -end_epoch "$epoch" -gpuid "$GPU_ID"
        touch "$DATA_PATH/model_part${part}_epoch${epoch}_1.23.t7"
      else
        th train.lua -data $DATA_PATH/data-$part-train.t7 -train_from "$prev_model" -save_model $DATA_PATH/model_part${part} -start_epoch "$epoch" -end_epoch "$epoch" -gpuid "$GPU_ID" >> "$LOG_PATH" 2>&1
      fi

    fi
  done
done


